"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JSTranslator = exports.makeNat = exports.makeBigNat = void 0;
/* eslint-disable no-case-declarations */
const DecisionTree_1 = require("../Compiler/DecisionTrees/DecisionTree");
const DecisionTreeCompiler_1 = require("../Compiler/DecisionTrees/DecisionTreeCompiler");
const DecisionTreeTranslator_1 = require("../Compiler/DecisionTrees/DecisionTreeTranslator");
const Utils_1 = require("../Compiler/Utils");
const Types_1 = require("../Parser/Types");
const Utils_2 = require("../Parser/Utils");
const Translator_1 = require("./Translator");
exports.makeBigNat = (nat) => `${nat}n`;
exports.makeNat = (nat) => `${nat}`;
function isFunCall(val) {
    return typeof val === 'object' && Types_1.dictHas(val, 'funName');
}
class JSTranslator extends DecisionTreeTranslator_1.DecisionTreeTranslator {
    constructor(trs, externals, nat = exports.makeBigNat) {
        super(trs, externals);
        this.nat = nat;
    }
    accessSubterm(parent, childIndex) {
        return `${parent}.args[${childIndex}]`;
    }
    init() {
        this.header.push(Utils_2.format('function isFun(term) {', 'return typeof term === "object";', '}'), Utils_2.format('function isVar(term) {', 'return typeof term === "string";', '}'), Utils_2.format('function isNat(term) {', 'const t = typeof term;', 'return t === "number" || t === "bigint";', '}'));
    }
    declareNullary(varName, symb) {
        return `const ${varName} = { name: "${symb}", args: [] };`;
    }
    callTerm(term) {
        if (Utils_1.isVar(term))
            return term;
        // if this is a free constructor simply output a term
        if (!this.isDefined(term.name)) {
            return this.translateTerm(Utils_1.fun(term.name, ...term.args.map(t => this.translateJSExpr(this.callTerm(t)))));
        }
        // otherwise call the corresponding js function
        const args = term.args.map(t => this.translateJSExpr(this.callTerm(t)));
        return { funName: term.name, args };
    }
    callOccTerm(occ, varNames) {
        if (DecisionTree_1.isOccurence(occ))
            return this.translateOccurence(occ, varNames);
        // if this is a free constructor simply output a term
        if (!this.isDefined(occ.name)) {
            return this.translateTerm(Utils_1.fun(occ.name, ...occ.args.map(t => this.translateJSExpr(this.callOccTerm(t, varNames)))));
        }
        // otherwise call the corresponding js function
        const args = occ.args.map(t => this.translateJSExpr(this.callOccTerm(t, varNames)));
        return { funName: occ.name, args };
    }
    translateOccurence(occ, varNames) {
        if (DecisionTree_1.isOccurence(occ)) {
            if (occ.pos.length === 0)
                return varNames[occ.index];
            return occ.pos.reduce((p, i) => `${p}.args[${i}]`, varNames[occ.index]);
        }
        if (occ.args.length === 0)
            return occ.name;
        return `${occ.name} (${occ.args.map(o => this.translateOccurence(o, varNames)).join(', ')})`;
    }
    translateDecisionTree(name, dt, varNames) {
        let hasTailRecursiveLeaf = false;
        const translate = (tree) => {
            switch (tree.type) {
                case 'fail':
                    return 'return ' + this.translateTerm(Utils_1.fun(name, ...varNames)) + ';';
                case 'leaf':
                    const ret = this.callOccTerm(tree.action, varNames);
                    // check if this is a tail recursive call
                    if (isFunCall(ret) && ret.funName === name) {
                        const newArgs = ret.args.map(this.translateJSExpr);
                        const updateArgs = Utils_2.map(Utils_1.zip(varNames, newArgs), ([v, newVal]) => `const ${v}_ = ${newVal};`);
                        const copyArgs = varNames.map(v => `${v} = ${v}_;`);
                        hasTailRecursiveLeaf = true;
                        return Utils_2.indent(1, '{', ...updateArgs, ...copyArgs, 'continue;', '}');
                    }
                    return `return ${this.translateJSExpr(ret)};`;
                case 'switch':
                    const cases = [];
                    for (const [ctor, A] of tree.tests) {
                        cases.push(Utils_2.indent(0, (ctor === DecisionTreeCompiler_1._ ? 'default:' : `case ${Translator_1.isNat(ctor) ? this.nat(ctor) : `"${ctor}"`}:`), Utils_2.indent(1, translate(A))));
                    }
                    const occName = this.translateOccurence(tree.occurence, varNames);
                    return Utils_2.indent(1, `switch (isFun(${occName}) ? ${occName}.name : ${occName}) {`, Utils_2.indent(1, cases.join('\n')), '}');
            }
        };
        const tree = translate(dt);
        const body = hasTailRecursiveLeaf ?
            `while(true) {\n ${tree} \n}` :
            tree;
        return Utils_2.indent(0, 'function ' + name + '(' + varNames.join(', ') + ') {', body, '}');
    }
    translateTerm(term) {
        if (Utils_1.isVar(term))
            return term;
        if (Translator_1.isNat(term.name))
            return this.nat(term.name);
        if (term.args.length === 0 && this.nullaries.has(term.name)) {
            return Translator_1.nullaryVarName(term.name);
        }
        return `{ name: "${term.name}", args: [${term.args.map(this.translateTerm).join(', ')}] }`;
    }
    translateJSExpr(expr) {
        if (typeof expr === 'string')
            return expr;
        if (isFunCall(expr)) {
            return `${expr.funName}(${expr.args.map(this.translateJSExpr).join(', ')})`;
        }
        return this.translateTerm(expr);
    }
}
exports.JSTranslator = JSTranslator;
